#!/usr/bin/env python3

# Author: Matthaeus Wander <mail@wander.science>

# Dependencies
import dns
import dns.ipv4
import dns.ipv6
import dns.message

import logging
import os
import socket
import struct
import sys

MAGIC = 0x00DABDAB

def file_parser(filenames):
    for filename in filenames:
        logging.info("Reading file {0}...".format(filename))

        with open(filename, 'rb') as fr:
            buf = fr.read(4)
            assert len(buf) == 4, "file too short, not a DNS file?"
            (magic,) = struct.unpack('I', buf)
            assert magic == MAGIC, "magic value does not match, not a DNS file?"

            message_count = 0
            file_size = os.stat(filename).st_size / 1024/1024

            eof = False
            while not eof:
                buf = fr.read(8)
                if len(buf) != 8:
                    eof = True
                    break

                (ts,) = struct.unpack('d', buf)

                buf = fr.read(5)
                (frameno, ipver) = struct.unpack('<IB', buf)

                if ipver == 4:
                    buf = fr.read(8)
                    ip_src = dns.ipv4.inet_ntoa(buf[0:4])
                    assert ip_src == socket.inet_ntoa(buf[0:4])
                    ip_dst = dns.ipv4.inet_ntoa(buf[4:8])
                    assert ip_dst == socket.inet_ntoa(buf[4:8])

                elif ipver == 6:
                    buf = fr.read(32)
                    ip_src = dns.ipv6.inet_ntoa(buf[0:16])
                    assert ip_src == socket.inet_ntop(socket.AF_INET6, buf[0:16])
                    ip_dst = dns.ipv6.inet_ntoa(buf[16:32])
                    assert ip_dst == socket.inet_ntop(socket.AF_INET6, buf[16:32])

                buf = fr.read(1 + 2 + 2 + 2)
                (transport_type, sport, dport, msglen) = struct.unpack('<BHHH', buf)
                msgbuf = fr.read(msglen)

                message_count += 1
                if message_count >= 5000:
                    message_count = 0
                    logging.info("Read {0}/{1} MBytes".format(fr.tell()/1024/1024, file_size))

                #print(ts, frameno, ipver, ip_src, ip_dst, transport_type, sport, dport, msglen)
                yield msgbuf, ts, frameno, ipver, ip_src, ip_dst, transport_type, sport, dport

def dns_parser(filenames):
    parser = file_parser(filenames)

    count = 0
    for (msgbuf, ts, frameno, ipver, ip_src, ip_dst, transport_type, sport, dport) in parser:
        try:
            dnsmsg = dns.message.from_wire(msgbuf, ignore_trailing=True)
            
            count += 1
            logging.debug("#{0} Processing DNS message of {1} bytes".format(count, len(msgbuf)))
            yield dnsmsg, len(msgbuf), ts, frameno, ipver, ip_src, ip_dst, transport_type, sport, dport

        # TODO: handle dns.message.ShortHeader        
        except dns.exception.FormError as e:
            logging.warning("Ignoring {0} bytes due to DNS parsing error: {1}".format(len(msgbuf), str(type(e))))
            continue

if __name__ == '__main__':
    import argparse
    arg = argparse.ArgumentParser(description='read DNS messages from .dns file')
    arg.add_argument('--debug', '-d', action='store_true', help='print debug log level')
    arg.add_argument('in_files', nargs='+', help='input dns filename(s)')
    args = arg.parse_args()

    loglevel = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(format='%(levelname)s %(message)s', level=loglevel)
    
    parser = dns_parser(args.in_files)

    for (dnsmsg, caplength, timestamp, frameno, ipver, ip_src, ip_dst, transport_type, sport, dport) in parser:
        print(dnsmsg.to_text())
