#!/usr/bin/env python3

# Author: Matthaeus Wander <mail@wander.science>

# Dependencies
import dpkt

import argparse
import bisect
import bz2
import itertools
import logging
import os
import socket
import struct
import sys

# config

IGNORE_ETHERTYPES = [
    dpkt.ethernet.ETH_TYPE_ARP,
]

MAGIC = 0x00DABDAB

CLEANUP_TIMEOUT = 600
CLEANUP_INTERVAL = 10000

# end of config

class MessageInfo(object):
    def __init__(self, frameno, packet_type, transport_type, ip_src, ip_dst):
        self.frameno = frameno
        self.packet_type = packet_type
        self.transport_type = transport_type
        self.ip_src = ip_src
        self.ip_dst = ip_dst

        self.sport = None
        self.dport = None

    def __repr__(self):
        return "MessageInfo(packet_type={0}, transport_type={1}, ip_src={2}, sport={3}, ip_dst={4}, dport={5})".format(
            self.packet_type, self.transport_type, self.ip_src, self.sport, self.ip_dst, self.dport)

    def is_ipv4(self):
        return self.packet_type == dpkt.ethernet.ETH_TYPE_IP

    def is_udp(self):
        return self.transport_type == dpkt.ip.IP_PROTO_UDP

class Reassembly(dict):
    def __init__(self):
        self.counter = 0

    def cleanup(self, ts):
        self.counter += 1
        if self.counter >= CLEANUP_INTERVAL:
            self.counter = 0
            logging.debug("Performing cleanup of stale reassembly buffers")
            self._cleanup(ts)

    def _cleanup(self, ts):
        remove_keys = []

        for key in self:
            diff = ts - self[key].time()
            if diff > CLEANUP_TIMEOUT:
                logging.debug("Deleting reassembly buffer \"{0}\" after {1}s timeout".format(key, diff))
                remove_keys.append(key)

        for key in remove_keys:
            del self[key]

class ReassemblyBuffer(object):
    def time(self):
        raise NotImplementedError()

###############
### File reader

def file_reader(filenames):
    for filename in filenames:
        if filename.endswith('.bz2'):
            logging.info("Opening file {} (BZ2)".format(filename))
            with bz2.BZ2File(filename, 'rb') as infile:
                yield infile, filename
        else:
            logging.info("Opening file {}".format(filename))
            with open(filename, 'rb') as infile:
                yield infile, filename

####################
### PCAP/Ethernet/IP

def ipv4_flag_df(ip):
    return (ip.off >> 14) & 0x1

def ipv4_flag_mf(ip):
    return (ip.off >> 13) & 0x1

def ipv4_offset(ip):
    return (ip.off & dpkt.ip.IP_OFFMASK) << 3

def pcap_parser(filenames):
    parser = file_reader(filenames)

    for infile, filename in parser:
        pcap_reader = dpkt.pcap.Reader(infile)
        link_type = pcap_reader.datalink()

        frameno = 0
        for ts, data in pcap_reader:
            frameno += 1
            logging.debug("PCAP frame no. {} file {}".format(frameno, filename))

            if link_type == dpkt.pcap.DLT_EN10MB: # Ethernet
                frame = dpkt.ethernet.Ethernet(data)
                packet_type = frame.type
            elif link_type == dpkt.pcap.DLT_LINUX_SLL: # Linux cooked capture (SLL)
                frame = dpkt.sll.SLL(data)
                packet_type = frame.ethtype
            else:
                raise Exception("Linktype " + str(link_type) + " is not implemented yet. Please report this error.")
            
            if packet_type not in (dpkt.ethernet.ETH_TYPE_IP, dpkt.ethernet.ETH_TYPE_IP6):
                if packet_type not in IGNORE_ETHERTYPES:
                    raise Exception("Warning: unknown EtherType", hex(packet_type) + ". Are you using this network protocol or did the parser fail to read ethernet frames?")
            
            ip = frame.data
            frag_info = None
            if packet_type == dpkt.ethernet.ETH_TYPE_IP:
                assert type(ip) is dpkt.ip.IP
                
                src = socket.inet_ntoa(ip.src)
                dst = socket.inet_ntoa(ip.dst)
                
                # Extract IPv4 fragmentation info
                if ipv4_flag_mf(ip) or ipv4_offset(ip) > 0:
                    frag_info = (ip.id, ipv4_offset(ip), ipv4_flag_mf(ip))
                    logging.debug("Encountered IPv4 fragment: {0} (id={1}) to {2}".format(src, ip.id, dst))

                yield ts, ip.data, MessageInfo(frameno, packet_type, ip.p, src, dst), frag_info

            elif packet_type == dpkt.ethernet.ETH_TYPE_IP6:
                assert type(ip) is dpkt.ip6.IP6

                src = socket.inet_ntop(socket.AF_INET6, ip.src)
                dst = socket.inet_ntop(socket.AF_INET6, ip.dst)

                # Extract IPv6 fragmentation info
                if dpkt.ip.IP_PROTO_FRAGMENT in ip.extension_hdrs:
                    ext = ip.extension_hdrs[dpkt.ip.IP_PROTO_FRAGMENT]
                    if ext is not None:
                        frag_info = (ext.id, ext.frag_off*8, ext.m_flag)
                        logging.debug("Encountered IPv6 fragment: {0} (id={1}) to {2}".format(src, ext.id, dst))

                # iterate through all IPv6 extension headers until transport header
                nxt = ip.nxt
                while nxt in ip.extension_hdrs:
                    nxt = ip.extension_hdrs[nxt].nxt

                yield ts, ip.data, MessageInfo(frameno, packet_type, nxt, src, dst), frag_info

####################
### IP fragmentation

class IpFragments(ReassemblyBuffer):
    def __init__(self, ts):
        self.buflist = []
        self.expected_size = None

        self.max_time = ts

    def append(self, offset, ts, data, last=False):
        for o, d in self.buflist:
            if o == offset:
                logging.debug("Skipping duplicate IP fragment")
                if len(d) != len(data):
                    logging.warning("Duplicate IP fragments have different lengths {0} != {1}".format(len(d), len(data)))
                return

        self.max_time = ts

        bisect.insort(self.buflist, (offset, data))

        if last:
            self.expected_size = offset + len(data)

    def finished(self):
        # expected size == actual size
        return self.expected_size is not None and self.expected_size == sum(map(lambda tup: len(tup[1]), self.buflist))

    def get_data(self):
        cl = None
        bufs = []
        for offset, data in self.buflist:
            if cl is None:
                cl = type(data)
            else:
                assert cl == type(data)

            # Convert dpkt.udp.UDP or dpkt.tcp.TCP to bytes
            bufs.append(bytes(data))

        # Return max_time when the packet was reassembled
        # Append buffers and convert back to UDP/TCP object
        return self.max_time, cl(b''.join(bufs))

    def time(self):
        return self.max_time

def fragmentation_parser(filenames):
    parser = pcap_parser(filenames)
    reassembly = Reassembly()

    for (ts, data, msg_info, frag_info) in parser:
        if frag_info is None:
            yield ts, data, msg_info
        else:
            (ip_id, frag_offset, frag_moar) = frag_info

            key = "ipfrag {0} id={1} {2}".format(msg_info.ip_src, ip_id, msg_info.ip_dst)
            if key not in reassembly:
                reassembly[key] = IpFragments(ts)

            reassembly[key].append(frag_offset, ts, data, last=not frag_moar)

            if reassembly[key].finished():
                (ts, data) = reassembly[key].get_data()
                logging.debug("Reassembled IP fragments: {0} bytes from {1}".format(len(data), key))
                del reassembly[key]
                yield ts, data, msg_info

            # remove stale reassembly buffers
            reassembly.cleanup(ts)

############################################
### TCP reassembly (order and deduplication)

# TCP sequence number arithmetic
def seq_normalize(a, b):
    boundary = (b + 2**31) % 2**32
    a = (a - boundary) % 2**32
    b = (b - boundary) % 2**32
    return a, b

def seq_greater(a, b):
    (a, b) = seq_normalize(a, b)
    return a > b

def seq_less_equal(a, b):
    (a, b) = seq_normalize(a, b)
    return a <= b

def seq_less(a, b):
    (a, b) = seq_normalize(a, b)
    return a < b

# Tests:
# assert seq_less_equal(2**32-10, 5) is True
# assert seq_less_equal(2**32-10, 2**31+10) is False

class TcpClose(object):
    pass

class TcpChannel(ReassemblyBuffer):
    def __init__(self, isn, ts):
        self.last_time = ts
        self.isn = isn
        self.seq = isn + 1
        self.buflist = []

    def put(self, bufseq, buftime, bufdata):
        assert buftime is not None
        self.last_time = buftime

        if type(bufdata) is TcpClose:
            if seq_less(bufseq, self.isn):
                logging.debug("Skipping misplaced TcpClose (maybe spoofed). SEQ={0} ISN={1}".format(bufseq, self.isn))
                return

        else:
            # skip old data, maybe dup of already processed data
            if seq_less_equal(bufseq + len(bufdata), self.seq):
                logging.debug("Skipping out-of-order TCP segment (maybe dup). BUFSEQ={0} BUFLEN={1}, expected SEQ={2}".format(bufseq, len(bufdata), self.seq))
                return

        # append *right* to prefer segments that arrived first
        self.buflist.append((bufseq, buftime, bufdata))

    def pop(self):
        index = 0
        while index < len(self.buflist):
            (bufseq, buftime, bufdata) = self.buflist[index]

            # old data, delete this segment
            if type(bufdata) is not TcpClose and seq_less_equal(bufseq + len(bufdata), self.seq):
                logging.debug("Skipping duplicate TCP segment. BUFSEQ={0} BUFLEN={1}, expected SEQ={2}".format(bufseq, len(bufdata), self.seq))
                self.buflist.pop(index)
                continue

            # future data, missing gap
            if seq_greater(bufseq, self.seq):
                index += 1
                continue

            # consecutive segment or close indication
            assert seq_less_equal(bufseq, self.seq)
            self.buflist.pop(index)

            if type(bufdata) is TcpClose:
                return buftime, bufdata

            else:
                assert seq_greater(bufseq + len(bufdata), self.seq)
                offset = self.seq - bufseq
                self.seq = (self.seq + len(bufdata[offset:])) % 2**32
                return (buftime, bufdata[offset:])

        # did not find matching data
        return (None, None)

    def time(self):
        return self.last_time

def transport_parser(filenames):
    parser = fragmentation_parser(filenames)
    channels = Reassembly()

    for (ts, transport, msg_info) in parser:
        if msg_info.transport_type == dpkt.ip.IP_PROTO_UDP:
            msg_info.sport = transport.sport
            msg_info.dport = transport.dport
            payload = transport.data
            if len(payload) == 0:
                logging.debug("Ignoring zero-length UDP datagram: ts={} {}".format(ts, msg_info))
                continue
            yield ts, payload, msg_info

        elif msg_info.transport_type == dpkt.ip.IP_PROTO_TCP:
            msg_info.sport = transport.sport
            msg_info.dport = transport.dport
            key = "tcp {0} p={1} {2} p={3}".format(msg_info.ip_src, transport.sport, msg_info.ip_dst, transport.dport)

            # new connection indicated by SYN
            if transport.flags & dpkt.tcp.TH_SYN:
                if key not in channels:
                    logging.debug("Opening TCP channel {0}".format(key))
                    channels[key] = TcpChannel(transport.seq, ts)

            if key not in channels:
                logging.debug("Unexpected TCP segment {0} (maybe dup)".format(key))
                continue

            # data
            if len(transport.data) > 0:
                logging.debug("TCP data for {0}".format(key))
                channels[key].put(transport.seq, ts, transport.data)

            # closing connection indicated by FIN or RST
            if (transport.flags & dpkt.tcp.TH_FIN) or (transport.flags & dpkt.tcp.TH_RST):
                channels[key].put(transport.seq + len(transport.data), ts, TcpClose())

            (buftime, bufdata) = channels[key].pop()
            while bufdata is not None:
                # TCP data
                if type(bufdata) is bytes:
                    yield buftime, bufdata, msg_info

                # close connection
                elif type(bufdata) is TcpClose:
                    logging.debug("Closing TCP channel {0}".format(key))
                    del channels[key]
                    break

                else:
                    assert False, "unreachable code"

                (buftime, bufdata) = channels[key].pop()

            # remove stale connection buffers
            channels.cleanup(ts)

        else:
            logging.warning("Unexpected transport type {} from {}".format(msg_info.transport_type, msg_info))

################################
### DNS message reassembly (TCP)

class DnsMessageBuffer(ReassemblyBuffer):
    LENGTH = 1
    MESSAGE = 2

    def __init__(self, ts):
        self.expecting = DnsMessageBuffer.LENGTH
        self.expected_size = 2

        self.actual_size = 0
        self.buflist = []

        self.last_time = ts

    def __len__(self):
        return self.actual_size

    def append(self, ts, buf):
        self.buflist.append((ts, buf))
        self.actual_size += len(buf)
        self.last_time = ts

    def get_bytes_left(self):
        return self.expected_size - self.actual_size

    def get_data(self):
        assert self.expected_size == self.actual_size
        bufs = []
        for ts, data in self.buflist:
            bufs.append(data)

        return ts, b''.join(bufs)

    def set_expected(self, expected_size):
        assert self.expecting == DnsMessageBuffer.LENGTH
        self.expected_size = expected_size
        self.actual_size = 0
        self.buflist = []
        self.expecting = DnsMessageBuffer.MESSAGE

    def time(self):
        return self.last_time

def message_parser(filenames):
    parser = transport_parser(filenames)
    reassembly = Reassembly()

    for (ts, payload, msg_info) in parser:
        if msg_info.transport_type == dpkt.ip.IP_PROTO_UDP:
            logging.info("Got {} bytes UDP DNS message from {}".format(len(payload), msg_info.ip_src))
            yield ts, payload, msg_info

        elif msg_info.transport_type == dpkt.ip.IP_PROTO_TCP:
            assert len(payload) > 0, "Payload in message parser is unexpectedly zero"

            buf_key = "tcpdns {0} p={1} {2} p={3}".format(msg_info.ip_src, msg_info.sport, msg_info.ip_dst, msg_info.dport)
            offset = 0
            while offset < len(payload):
                payload_bytes = len(payload)-offset

                if buf_key not in reassembly:
                    reassembly[buf_key] = DnsMessageBuffer(ts)

                buf = reassembly[buf_key]
                bytes_left = buf.get_bytes_left()

                # message finished
                if bytes_left <= payload_bytes:
                    buf.append(ts, payload[offset:offset+bytes_left])
                    (ts, msg) = buf.get_data()
                    offset += bytes_left

                    if buf.expecting == DnsMessageBuffer.LENGTH:
                        (msg_size,) = struct.unpack('>H', msg)
                        logging.debug("Attempting to read {} bytes TCP dns message".format(msg_size))
                        buf.set_expected(msg_size)
                        continue

                    elif buf.expecting == DnsMessageBuffer.MESSAGE:
                        logging.info("Got {} bytes TCP DNS message from {}".format(buf.expected_size, msg_info.ip_src))
                        del reassembly[buf_key]
                        yield ts, msg, msg_info
                        continue

                    assert False, "unreachable"

                # still more to fetch
                else:
                    buf.append(ts, payload[offset:len(payload)])
                    logging.debug("Reassembly in progress {0}/{1} bytes".format(buf.actual_size, buf.expected_size))
                    break

            # remove stale message buffers
            reassembly.cleanup(ts)

        else:
            #print("Encountered unknown payload type within IPv4: ", msg_info.transport_type)
            raise Exception("Encountered unknown payload type within IPv4: ", msg_info.transport_type)

if __name__ == '__main__':
    arg = argparse.ArgumentParser(description='reassemble DNS messages from pcap files')
    arg.add_argument('--debug', '-d', action='store_true', help='print debug log level')
    arg.add_argument('in_files', nargs='+', help='input pcap filename(s)')
    arg.add_argument('out_file', help='output filename')
    args = arg.parse_args()

    loglevel = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(format='%(levelname)s %(message)s', level=loglevel)

    with open(args.out_file, 'wb') as fw:
        fw.write(struct.pack('<I', MAGIC))

        parser = message_parser(args.in_files)
        for ts, msgbuf, msg_info in parser:
            fw.write(struct.pack('d', ts))
            fw.write(struct.pack('<I', msg_info.frameno))

            if msg_info.is_ipv4():
                fw.write(struct.pack('<B', 4))
                fw.write(socket.inet_aton(msg_info.ip_src))
                fw.write(socket.inet_aton(msg_info.ip_dst))
                
            else:
                fw.write(struct.pack('<B', 6))
                fw.write(socket.inet_pton(socket.AF_INET6, msg_info.ip_src))
                fw.write(socket.inet_pton(socket.AF_INET6, msg_info.ip_dst))

            fw.write(struct.pack('<B', msg_info.transport_type))
            fw.write(struct.pack('<H', msg_info.sport))
            fw.write(struct.pack('<H', msg_info.dport))

            fw.write(struct.pack('<H', len(msgbuf)))
            fw.write(msgbuf)
